package generateVoyages;

import java.util.ArrayList;

import master.DataReader;
import master.DeliveryNode;
import master.Node;
import master.PickupNode;

public class VoyageGenerator {
	
	/**
	 * Takes a DataReader instance as input and creates all feasible, non-dominated voyages for each vessel in the instance through n stages, where n is the number of nodes in the instance.
	 * @param input
	 * @return
	 */
	public ArrayList<ArrayList<State>> findAllVoyages(DataReader input){
		
		int[] capacity = input.getCapacity();
		int[] minTime = input.getMinTime();
		double[] maxTime = input.getMaxTime();
		
		ArrayList<ArrayList<State>> finalStates = new ArrayList<ArrayList<State>>(capacity.length); //All optimal paths starting and ending in the supply depot
		
		ArrayList<ArrayList<State>> states = new ArrayList<ArrayList<State>>(); //List containing all stages
		ArrayList<State> stage0 = new ArrayList<State>(); //List containing all states at stage 0 (that is, only the initial state)
		
		ArrayList<State> dominated = new ArrayList<State>();
		
		ArrayList<Node> nodes = input.getNodeList();
		ArrayList<PickupNode> pickupNodes = input.getPickupNodeList();
		ArrayList<DeliveryNode> deliveryNodes = input.getDeliveryNodeList();
		Node depotDelivery = deliveryNodes.get(0);
		
		State initState = new State(pickupNodes.get(0), capacity, 0.0, 0.0, minTime, maxTime); //Initial state
		stage0.add(initState);
		states.add(stage0);
		
		State tempState = null;
		
		int counter = 0;
		int i = 0;
		while(i < nodes.size()-1){
			states.add(new ArrayList<State>());
			for(State state:states.get(i)){
				for(Node node:nodes){
					if(((state.getVisited() % Math.pow(2, node.getNodeNumber()+1)) >= Math.pow(2, node.getNodeNumber()) || //Hvis stien allerede inneholder noden; continue
							state.getCurrNode() == depotDelivery)){ //Hvis stien ender i depot-delivery; continue
						continue; 
					}
					if(node.getSibling() != null && node != depotDelivery && node instanceof DeliveryNode){//Pickup node should never be visited before delivery node for the same installation 
						if(state.getVisited() % Math.pow(2, node.getSibling().getNodeNumber()+1) >= Math.pow(2, node.getSibling().getNodeNumber())){
//							System.out.println("No pickup before delivery");
							continue;
						}
					}
					if(node == depotDelivery){//Extends the state to the depot delivery node
						if(state.getCurrNode() == pickupNodes.get(0)) continue;
						tempState = new State(state, node, 
										state.getBackload(), 
										state.getDemand(), 
										state.getCost() + state.getCurrNode().getServiceSailingCosts(node),
										state.getCurrTime() + state.getCurrNode().getSailingTime(node));
										tempState.setVisited(state.getVisited());
										tempState.setVisited(node);
//						System.out.println("Ends in depot");
					}
					else if(node instanceof PickupNode){//Extends the state to a pickup node
						tempState = new State(state, node, 
										state.getBackload() + ((PickupNode) node).getBackload(), 
										Math.max(state.getDemand(), state.getBackload() + ((PickupNode) node).getBackload()), 
										state.getCost() + state.getCurrNode().getServiceSailingCosts(node),
										state.getCurrTime() + state.getCurrNode().getSailingTime(node));
										tempState.setVisited(state.getVisited());
										tempState.setVisited(node);
//						System.out.println("New pickup");
					}
					else{//Extends the state to a delivery node
						tempState = new State(state, node, 
										state.getBackload(), 
										Math.max(state.getDemand() + ((DeliveryNode)node).getDemand(), state.getBackload()), 
										state.getCost() + state.getCurrNode().getServiceSailingCosts(node),
										state.getCurrTime() + state.getCurrNode().getSailingTime(node));
										tempState.setVisited(state.getVisited());
										tempState.setVisited(node);
//						System.out.println("New delivery");
					}
					
					for(int j=0; j<capacity.length; j++){
						if(feasibleMove(tempState, depotDelivery)[j]){ //Add tempState to the new stage if tempState is feasible for one of the vessels
							try {
								states.get(i+1).add( (State) tempState.clone());
	//							System.out.println("New state");
							} catch (CloneNotSupportedException e) {
								e.printStackTrace();
							}
							//Creates list of dominated states
							dominated = dominatesList(states.get(i+1).get(states.get(i+1).size()-1), states.get(i+1), depotDelivery, dominated);
							break;
						}
					}
//					else System.out.println("Move not feasible");
				}
			}
			
			i++;
			
			System.out.println("Number of states in stage " + i + ": " + states.get(i).size());
			
			//Removes dominated states from the new stage
			for(int j = 0; j < dominated.size(); j++){
				states.get(i).remove(dominated.get(j));
			}
			dominated.clear();

			//All feasible, not dominated states, ending in the depot delivery node are added to the list of final states
			counter = 0;
			boolean[] feasibility;
			for(int j = 0; j<states.get(i).size(); j++){
				if(states.get(i).get(j).getCurrNode() == depotDelivery){
					feasibility = feasibleMove(states.get(i).get(j), depotDelivery);
					for(int k = 0; k<capacity.length; k++){
						if(feasibility[k]){
							if(finalStates.size()<capacity.length){
								finalStates.add(new ArrayList<State>());
							}
							finalStates.get(k).add(states.get(i).get(j));
						}
					}
					counter++;
				}
			}
			System.out.println("Number of possible voyages of length " + (i + 1) + ": " + counter);
		}
		return finalStates;
	}
	
	/**
	 * Takes a state and the depot delivery node as input and checks whether the state is feasible with respect to time and capacity restrictions.
	 * @param state
	 * @param depotDelivery
	 * @return
	 */
	//Checks whether a state is feasible
	private static boolean[] feasibleMove(State state, Node depotDelivery){
		boolean[] feasibility = new boolean[state.getCapacity().length];
		for(int i=0; i<state.getAvailableTime().length; i++){
			if(state.getCurrNode() == depotDelivery && state.getCurrTime() > state.getAvailableTime()[i]){
				feasibility[i] = false;
				continue;
			}
			else if(state.getCurrTime() + state.getCurrNode().getSailingTime(depotDelivery) > state.getAvailableTime()[i]){
				feasibility[i] = false;
				continue;
			}
			else if(state.getDemand() > state.getCapacity()[i]){
				feasibility[i] = false;
				continue;
			}
			else if(state.getBackload() > state.getCapacity()[i]){
				feasibility[i] = false;
				continue;
			}
			else{
				feasibility[i] = true;
			}
		}
		return feasibility;
	}
	
	//Returns the list dominated of states dominated by state
	/**
	 * The method takes a new state, the list of states in the current stage, the depot delivery node, and the list of dominated states in the current stage as input. It returns a list contaning the new state if it is dominated or a list containg the states that the new state dominates.
	 * @param state
	 * @param states
	 * @param depotDelivery
	 * @param dominated
	 * @return
	 */
	private static ArrayList<State> dominatesList(State state, ArrayList<State> states, Node depotDelivery, ArrayList<State> dominated){
		for(int i=0; i<states.size()-1; i++){
			
			//If the states do not visit or end in the same nodes, next iteration
			if(states.get(i).getVisited() != state.getVisited() || states.get(i).getCurrNode() != state.getCurrNode()) continue;
			
			//If both states end in the depot delivery node, only time and cost is relevant for domination
			if(state.getCurrNode()==depotDelivery && states.get(i).getCurrNode() == depotDelivery){
					if(state.getCost()<=states.get(i).getCost() && state.getCurrTime()<=states.get(i).getCurrTime()){
						if(!dominated.contains(states.get(i))) 
						dominated.add(states.get(i));
//						System.out.println("Dominated 01");
						continue;
					}
					else if(state.getCost()>=states.get(i).getCost() && state.getCurrTime()>=states.get(i).getCurrTime()){
						dominated.add(state);
//						System.out.println("Dominated 02");
//						continue;
						return dominated;
					}
					else{
						continue;
					}
			}
			
			//If one of the states dominate the other, add the dominated state to dominated
			if(state.getCost()<=states.get(i).getCost()){
				if(state.getCurrTime()>states.get(i).getCurrTime()){
					continue;
				}
				else if(state.getBackload()>states.get(i).getBackload()){
					continue;
				}
				else if(state.getDemand()>states.get(i).getDemand()){
					continue;
				}
				else{
					if(!dominated.contains(states.get(i))) 
					dominated.add(states.get(i));
//					System.out.println("Dominated 1");
				}
			}
			if(state.getCost()>=states.get(i).getCost()){
				if(state.getCurrTime()<states.get(i).getCurrTime()){
					continue;
				}
				else if(state.getBackload()<states.get(i).getBackload()){
					continue;
				}
				else if(state.getDemand()<states.get(i).getDemand()){
					continue;
				}
				else{
					dominated.add(state);
					return dominated;
//					System.out.println("Dominated 2");
				}
			}
		}
		return dominated;
	}
}
