package master;

import java.util.ArrayList;
import java.util.Random;
/**
 * Neighborhood moves used by the shaking and local search
 * @author Nils Albjerk
 * @author Teodor Danielsen
 * @author Stian Krey
 *
 */
public class Moves {
	private Random rand = new Random();
	/**
	 * Takes two random nodes from the route, removes one edge from each of these that are not neighboring edges, 
	 * reconnects the graph, and switches the direction of visit of the intermediate nodes.
	 * @param vessel the vessel the operator is performed on
	 * @return a new route
	 */
	public ArrayList<Node> twoOpt(Vessel vessel){
	
		ArrayList<Node>temp = vessel.cloneRoute();
		ArrayList<Node>new_route = new ArrayList<Node>();
		int size = temp.size();
		int i = 0;
		int k = 0;
		int min = 1;
		int max = size-2;
		while (i == k && i>k){
			i = rand.nextInt(max-min+1)+min;
			k = rand.nextInt(max-min+1)+min;
		}
		
		for (int j = 0; j < i; j++) {
			new_route.add(temp.get(j));
		}
		
		for (int l = k; l>=i; l--){
			new_route.add(temp.get(l));
		}
		
		for (int m = k+1; m<temp.size(); m++){
			new_route.add(temp.get(m));
		}
		
		return new_route;
	}
	
	/**
	 * Three random nodes are selected from the route, one non-neighboring edge is removed from each node, and  the graph is reconnected. 
	 * The intermediate nodes between two of the chosen nodes has their direction of visit reversed.
	 * @param vessel the vessel the operator is performed on
	 * @return a new route
	 */
	public ArrayList<Node> threeOpt(Vessel vessel){
		ArrayList<Node>temp = vessel.cloneRoute();
		ArrayList<Node>new_route = new ArrayList<Node>();
		int size = temp.size();
		if(size<=6){
			return temp;
		}
		int i = 0;
		int j = 0;
		int k = 0;
		int min = 1;
		int max = size-2;
		while (i == k || i == j || j == k || i>j || j>k){
			i = rand.nextInt(max-min+1)+min;
			j = rand.nextInt(max-min+1)+min;
			k = rand.nextInt(max-min+1)+min;	
		}
		for (int l = 0; l < i; l++){
			new_route.add(temp.get(l));
		}
		
		for (int l = k; l >= j; l--){
			new_route.add(temp.get(l));
		}
		
		for (int l = i; l < j; l++){
			new_route.add(temp.get(l));
		}
		
		for (int l = k+1; l < temp.size(); l++){
			new_route.add(temp.get(l));
		}
		
		return new_route;
	}
	
	/**
	 * Takes two random nodes in the route, and swaps their location in the sequence of visits.
	 * @param vessel the vessel the operator is performed on
	 * @return a new route
	 */
	public ArrayList<Node> swap(Vessel vessel){
		ArrayList<Node>temp = vessel.cloneRoute();
		int size = temp.size();
		if(size<=3){
			return temp;
		}
		
		int i = 0;
		int k = 0;
		int min = 1;
		int max = size-2;
		while(i==k){
			i = rand.nextInt(max-min+1)+min;
			k = rand.nextInt(max-min+1)+min;
		}
		
		Node nodeI = temp.get(i);
		Node nodeK = temp.get(k);
		temp.set(i, nodeK);
		temp.set(k, nodeI);
		return temp;
	}
	
	/**
	 * Takes a random node, removes it from the route, and inserts it in another random place in the same route.
	 * @param vessel the vessel the operator is performed on
	 * @return a new route
	 */
	public ArrayList<Node> insert(Vessel vessel){
		ArrayList<Node>temp = vessel.cloneRoute();
		int size = temp.size();
		if(size<=3){
			return temp;
		}

		int i = 0;
		int pos = 0;
		int min = 1;
		int max = size-2;
		while(i == pos){
			i = rand.nextInt(max-min+1)+min;
			pos = rand.nextInt(max-min+1)+min;
		}
		
		Node node = temp.get(i);
		temp.remove(i);
		temp.add(pos, node);
		return temp;
	}
	
	/**
	 * Takes a random number m of sequential nodes from one voyage and exchanges them with n nodes from the other voyage.
	 * n is randomly chosen to be either m or m-1.
	 * @param vessel1 the first vessel the operator is performed on
	 * @param vessel2 the second vessel the operator is performed on
	 * @return an ArrayList with two new routes
	 */
	public ArrayList<ArrayList<Node>> exchange(Vessel vessel1, Vessel vessel2){
		ArrayList<Node>temp1 = vessel1.cloneRoute();
		ArrayList<Node>temp2 = vessel2.cloneRoute();
		ArrayList<Node>new1 = new ArrayList<Node>();
		ArrayList<Node>new2 = new ArrayList<Node>();
		
		ArrayList<ArrayList<Node>> new_routes = new ArrayList<ArrayList<Node>>();
		
		int size1 = temp1.size();
		
		int size2 = temp2.size();
		
		if(size1<=3 || size2<=3){
			new_routes.add(temp1);
			new_routes.add(temp2);
			return new_routes;
		}

		int smallest;
		if(size1<=size2){
			smallest = size1;
		}else{
			smallest = size2;
		}
		int m = rand.nextInt(((smallest/2)-1)+1)+1;
		int n;
		
		if(rand.nextBoolean()){
			n=m;
		}else{
			if (m==1){
				n=m;
			}else{
				n=m-1;
			}
		}
		int startEx1 = rand.nextInt(((size1-m)-1)+1)+1;
		int startEx2 = rand.nextInt(((size2-n)-1)+1)+1; 
		for (int i = 0; i < startEx1; i++) {
			new1.add(temp1.get(i));
		}
		for (int i = startEx2; i < startEx2+n; i++) {
			new1.add(temp2.get(i));
		}
		for (int i = startEx1+m; i < temp1.size(); i++) {
			new1.add(temp1.get(i));
		}
		
		for (int i = 0; i < startEx2; i++) {
			new2.add(temp2.get(i));
		}
		for (int i = startEx1; i < startEx1+m; i++) {
			new2.add(temp1.get(i));
		}
		for (int i = startEx2+n; i < temp2.size(); i++) {
			new2.add(temp2.get(i));
		}
		
		new_routes.add(new1);
		new_routes.add(new2);
		
		return new_routes;
		
	}
	
	/**
	 * Deletes the edges between two consecutive nodes in each route, and reconnects the routes with each other.
	 * @param vessel1 the first vessel the operator is performed on
	 * @param vessel2 the second vessel the operator is performed on
	 * @return an ArrayList with two new routes
	 */
	public ArrayList<ArrayList<Node>> cross(Vessel vessel1, Vessel vessel2){
		ArrayList<Node>temp1 = vessel1.cloneRoute();
		ArrayList<Node>temp2 = vessel2.cloneRoute();
		
		int size1 = temp1.size();
		int size2 = temp2.size();
		ArrayList<ArrayList<Node>> new_routes = new ArrayList<ArrayList<Node>>();
		if(size1<=3 || size2<=3){
			new_routes.add(temp1);
			new_routes.add(temp2);
			return new_routes;
		}

		int startCross1 = rand.nextInt((size1/2)-1)+1;
		int lengthCross1 = rand.nextInt(size1-startCross1-2)+1;
		int startCross2 = rand.nextInt((size2/2)-1)+1;
		int lengthCross2 = rand.nextInt(size2-startCross2-2)+1;
		ArrayList<Node>new1 = new ArrayList<Node>();
		ArrayList<Node>new2 = new ArrayList<Node>();
		
		for (int i = 0; i < startCross1; i++) {
			new1.add(temp1.get(i));
		}
		for (int i = startCross2; i < startCross2+lengthCross2; i++) { //
			new1.add(temp2.get(i));
		}
		for (int i = startCross1+lengthCross1; i < temp1.size(); i++) {
			new1.add(temp1.get(i));
		}
		
		for (int i = 0; i < startCross2; i++) {
			new2.add(temp2.get(i));
		}
		for (int i = startCross1; i < startCross1+lengthCross1; i++){
			new2.add(temp1.get(i));
		}
		for (int i = startCross2+lengthCross2; i < temp2.size(); i++){
			new2.add(temp2.get(i));
		}
		
		new_routes.add(new1);
		new_routes.add(new2);
		
		return new_routes;
	}
	
	/**
	 * Removes a random node from the first route, and inserts it into a random location in the second route.
	 * @param vessel1 the first vessel the operator is performed on
	 * @param vessel2 the second vessel the operator is performed on
	 * @return an ArrayList with two new routes
	 */
	public ArrayList<ArrayList<Node>> shift(Vessel vessel1, Vessel vessel2){
		ArrayList<Node>temp1 = vessel1.cloneRoute();
		ArrayList<Node>temp2 = vessel2.cloneRoute();
		
		ArrayList<ArrayList<Node>> new_routes = new ArrayList<ArrayList<Node>>();
		if(temp1.size()<=2){
			new_routes.add(temp1);
			new_routes.add(temp2);
			return new_routes;
		}

		int fromPos = rand.nextInt(((temp1.size()-2)-1)+1)+1;
		int toPos = 0;
		if(temp2.size()<=2){
			toPos = 1;
		}else{
			toPos = rand.nextInt(((temp2.size()-2)-1)+1)+1;
		}
		Node node = temp1.get(fromPos);
		temp1.remove(fromPos);
		temp2.add(toPos, node);
		
		new_routes.add(temp1);
		new_routes.add(temp2);
		
		return new_routes;
	}
	
	/**
	 * Takes one node randomly from each route and exchanges their locations.
	 * @param vessel1 the first vessel the operator is performed on
	 * @param vessel2 the second vessel the operator is performed on
	 * @return an ArrayList with two new routes
	 */
	public ArrayList<ArrayList<Node>> replace(Vessel vessel1, Vessel vessel2){
		ArrayList<Node>temp1 = vessel1.cloneRoute();
		ArrayList<Node>temp2 = vessel2.cloneRoute();
		
		ArrayList<ArrayList<Node>> new_routes = new ArrayList<ArrayList<Node>>();
		if(temp1.size()<=3 || temp2.size()<=3){
			new_routes.add(temp1);
			new_routes.add(temp2);
			return new_routes;
		}

		int fromPos1 = rand.nextInt(((temp1.size()-2)-1)+1)+1;
		int fromPos2 = rand.nextInt(((temp2.size()-2)-1)+1)+1;
		Node node1 = temp1.get(fromPos1);
		Node node2 = temp2.get(fromPos2);
		temp1.remove(fromPos1);
		temp2.remove(fromPos2);
		temp1.add(fromPos1, node2);
		temp2.add(fromPos2, node1);
		
		new_routes.add(temp1);
		new_routes.add(temp2);
		
		return new_routes;
	}
}
