package master;

import java.util.ArrayList;

/**
 * Used to represent the offshore supply vessels (OSVs) in the problem
 * @author Nils Albjerk
 * @author Teodor Danielsen
 * @author Stian Krey
 *
 */
public class Vessel implements Cloneable {
	
	int vesselName;
	int capacity;
	double maxTime;
	double minTime;
	double loadCoef = 1;
	int MaxDelay = 24;
	
	boolean spot;
	double fixedCharterCost;
	int timeCharterCost;
	
	public ArrayList<Node>route;
	int postponeCost;
	int delayCost;

	public Object clone(){
		try {
			return super.clone();
		} catch (Exception e) {
			return null;
		}
	}
	
	public ArrayList<Node> cloneRoute(){
		ArrayList<Node> temp = new ArrayList<Node>();
		for (int i = 0; i < route.size(); i++) {
			temp.add(route.get(i));
		}
		return temp;
	}
	
	public Vessel(Vessel clone){
		this.vesselName = clone.vesselName;
		this.capacity = clone.capacity;
		this.maxTime = clone.maxTime;
		this.minTime = clone.minTime;
		
		if (clone.spot) {
			this.spot = true;
			this.fixedCharterCost = clone.fixedCharterCost;
			this.timeCharterCost = clone.timeCharterCost;
		}else{
			this.spot = false;
		}
		this.route = (ArrayList<Node>) clone.route.clone();
		this.postponeCost = clone.postponeCost;
		this.delayCost = clone.delayCost;
			
	}
	
	public Vessel (int capacity, double maxTime, double minTime, int vesselName, int postponeCost, int delayCost){
		this.vesselName = vesselName;
		this.capacity = capacity;
		this.maxTime = maxTime;
		this.minTime = minTime;
		this.route = new ArrayList<Node>();
		this.postponeCost = postponeCost;
		this.delayCost = delayCost;
		
	}
	
	public Vessel (int capacity, double maxTime, double minTime, double fixedCharterCost, int timeCharterCost, int vesselName, int postponeCost, int delayCost){
		this.vesselName = vesselName;
		this.capacity = capacity;
		this.maxTime = maxTime;
		this.minTime = minTime;
		this.fixedCharterCost = fixedCharterCost;
		this.timeCharterCost = timeCharterCost;
				this.route = new ArrayList<Node>();
				this.postponeCost = postponeCost;
				this.delayCost = delayCost;
	}
	
	public int getCapacity() {
		return capacity;
	}
	public void setCapacity(int capacity) {
		this.capacity = capacity;
	}
	public double getMaxTime() {
		return maxTime;
	}
	public void setMaxTime(double maxTime) {
		this.maxTime = maxTime;
	}
	public double getMinTime() {
		return minTime;
	}
	public void setMinTime(double minTime) {
		this.minTime = minTime;
	}
	
	/**
	 * Calculates the cost of using this vessel
	 * @return the cost of the vessel
	 */
	public double cost(){
		double cost = voyageCost();
		
		if(route.size() > 2){
			if(spot){
				cost += fixedCharterCost;
 
				cost += timeCharterCost*Math.ceil(voyageSailingTime()/24.0);
			}
		}
		return cost;
	}
	
	/**
	 * Checks if the vessels breaks the constraints regarding the total sailing time, and capacity on board during the route
	 * @return a boolean that states if the vessel's route is feasible
	 */
	public boolean vesselOK(){
		if (spot) {
			if(voyageCapOK() && voyageSailingTime()+minTime <= maxTime) // not spossible to have 24h delay
				return true;
			else
				return false;
		}
		else{
			if(voyageCapOK() && voyageSailingTime()+minTime <= maxTime + MaxDelay) //possible to have 24h delay
				return true;
			else
				return false;	
		}
		
	}
	
	public void printVoyage(){
		System.out.println("Vessel #: "+ vesselName);
		System.out.print("Route: ");
		for (int i = 0; i < route.size(); i++) {
			System.out.print(route.get(i).getNodeNumber()+", ");
		}
		System.out.print("Route: ");
		System.out.println("");for (int i = 0; i < route.size(); i++) {
			System.out.print(route.get(i).getNodeName()+", ");
		}
		System.out.println("");
	}
	
	public double voyageCost(){
		double cost = 0;
		for (int i = 0; i < route.size()-1; i++) {
			cost += route.get(i).getServiceSailingCosts(route.get(i+1));
		}
		if(voyageSailingTime()+minTime > maxTime){
			cost += (voyageSailingTime()+minTime - maxTime)*delayCost;
		}
//		cost += loadCost();
		return cost;
	}
	
	/**
	 * Calculates the cost of adding a node at a specific positions
	 * @param node the node to be added
	 * @param pos the position of the node
	 * @return the cost of adding the node to the route
	 */
	public double voyageCostAdd(Node node, int pos){
		double cost = 0;
		ArrayList<Node>routeCopy = new ArrayList<Node>();
		for (int i = 0; i < route.size(); i++) {
			routeCopy.add(route.get(i));
		}

		routeCopy.add(pos, node);
		for (int i = 0; i < routeCopy.size()-1; i++) {
			cost += routeCopy.get(i).getServiceSailingCosts(routeCopy.get(i+1));
		}
		if(voyageSailingTime()+minTime > maxTime){
			cost += (voyageSailingTime()+minTime - maxTime)*delayCost;
		}
		return cost;
	}
	
	/**
	 * NOT IN USE
	 * @return
	 */
	public double loadCost(){
		double cost = 0;
		double startLoad = 0;

		for (int i = 0; i < route.size(); i++) {
			if(route.get(i) instanceof DeliveryNode){
				DeliveryNode d = (DeliveryNode) route.get(i);
				startLoad = startLoad + d.getDemand(); 
			}	
		}

		double load = startLoad;
		//sjekk for alle steg i ruta.
		for (int i = 1; i < route.size(); i++) {	// starts with node 1 in the route,
			cost += route.get(i-1).getSailingTime((route.get(i)))*load*loadCoef;
			if(route.get(i) instanceof DeliveryNode){
				DeliveryNode d = (DeliveryNode) route.get(i);
				load = load - d.getDemand();
			}
			if(route.get(i) instanceof PickupNode){
				PickupNode p = (PickupNode) route.get(i);
				load = load + p.getBackload();
			}
			
		}
		return cost;
	}
	
	/**
	 * Calculates the time used to sail the vessel's route
	 * @return the sailing time for the route
	 */
	public double voyageSailingTime(){
		double time = 0;
		for (int i = 0; i < this.route.size()-1; i++) {
			time += this.route.get(i).getSailingTime(route.get(i+1));
		}
		return time;
	}
	
	/**
	 * Calculates the time used to sail the vessel's route if a node is added at the specific position
	 * @param index the position
	 * @param node the node to be added
	 * @return the sailing time when the new node is added to the route
	 */
	public boolean voyageSailingTimeAdd(int index, Node node){
		ArrayList<Node> temproute = new ArrayList<Node>();
		for (int i = 0; i < route.size(); i++) {
			temproute.add(route.get(i));
		}
		temproute.add(index, node);
		
		double time = 0;
		for (int i = 0; i < temproute.size()-1; i++) {
			time += temproute.get(i).getSailingTime(temproute.get(i+1));
		}
		if(vesselTimeOk(time)){
			return true;
		}
		return false;		
		
	}
	
	/**
	 * Calculates the time used to sail the vessel's route if a node is removed from the route
	 * @param index the position of the node to be removed
	 * @return the sialing time when the node is removed
	 */
	public double voyageSailingTimeRemove(int index){
		ArrayList<Node> tempRoute = (ArrayList<Node>) route.clone();
		Node node = tempRoute.get(index);
		if(tempRoute.get(index).getSibling()!=null){
			tempRoute.remove(tempRoute.get(index).getSibling());
		}
		tempRoute.remove(node);
		
		double time = 0;
		for (int i = 0; i < tempRoute.size()-1; i++) {
			time += tempRoute.get(i).getSailingTime(tempRoute.get(i+1));
		}
		return time;
	}
	
	/**
	 * Checks if the sailing time is within the allowable limits, including the maximum delay
	 * @param sailingTime the sailing time for the route
	 * @return a boolean stating if the sailing time is feasible
	 */
	public boolean vesselTimeOk(double sailingTime){
		if (spot) {
			if(sailingTime+minTime <= maxTime) // not spossible to have 24h delay
				return true;
			else
				return false;
		}
		else{
			if(sailingTime+minTime <= maxTime + MaxDelay) //possible to have 24h delay
				return true;
			else
				return false;	
		}
		
	}
	
	/**
	 * Checks if the capacity of the vessel is feasible for all legs of the route
	 * @return a boolean stating if the capacity is feasible
	 */
	public boolean voyageCapOK(){
		double startLoad = 0;

		for (int i = 0; i < route.size(); i++) {
			if(route.get(i) instanceof DeliveryNode){
				DeliveryNode d = (DeliveryNode) route.get(i);
				startLoad = startLoad + d.getDemand(); 
			
			}	
		}
		if(startLoad>capacity){
			return false;
		}
		
		double load = startLoad;
		
		//sjekk for alle steg i ruta.
		for (int i = 0; i < route.size(); i++) {	
			if(route.get(i) instanceof DeliveryNode){
				DeliveryNode d = (DeliveryNode) route.get(i);
				load = load - d.getDemand();
			}
			if(route.get(i) instanceof PickupNode){
				PickupNode p = (PickupNode) route.get(i);
				load = load + p.getBackload();
			}
			
			if(load > capacity){
				return false;
			}
		}
		
		return true;
		
	}
	
	//check for route when adding node
	public boolean voyageCapOKAdd(int index, Node node){
		ArrayList<Node> temproute = new ArrayList<Node>();
		for (int i = 0; i < route.size(); i++) {
			temproute.add(route.get(i));
		}
		temproute.add(index, node);
		double startLoad = 0;
		
		for (int i = 0; i < temproute.size(); i++) {
			if(temproute.get(i) instanceof DeliveryNode){
				DeliveryNode d = (DeliveryNode) temproute.get(i);
				startLoad = startLoad + d.getDemand(); 
			
			}	
		}
		
		if(startLoad>capacity)
			return false;
		
		double load = startLoad;
		
		//sjekk for alle steg i ruta.
		for (int i = 0; i < temproute.size(); i++) {	// starts with node 1 in the route,
			if(temproute.get(i) instanceof DeliveryNode){
				DeliveryNode d = (DeliveryNode) temproute.get(i);
				load = load - d.getDemand();
			}
			if(temproute.get(i) instanceof PickupNode){
				PickupNode p = (PickupNode) temproute.get(i);
				load = load + p.getBackload();
			}
			
			if(load > capacity)
				return false;
		}
		
		return true;
	}
	



	public Node getLastElement(){
		return this.route.get(this.route.size());
	}
	
	public double printLoadCost(){
		double cost = 0;
		double startLoad = 0;

		for (int i = 0; i < route.size(); i++) {
			if(route.get(i) instanceof DeliveryNode){
				DeliveryNode d = (DeliveryNode) route.get(i);
				startLoad = startLoad + d.getDemand(); 
			
			}	
		}
		System.out.println("StartLoad " + startLoad);

		double load = startLoad;
		System.out.println("Distance from node "+ route.get(0).getNodeNumber()+" to node  "+route.get(1).getNodeNumber()+ " : " + route.get(0).getSailingTime((route.get(1))));
		System.out.println("Start load cost: " + cost);
		//sjekk for alle steg i ruta.
		for (int i = 1; i < route.size(); i++) {	// starts with node 1 in the route,
			System.out.println("Load when in node  "+route.get(i).getNodeNumber()+ ": "+ load);
			System.out.println("Distance from node " + route.get(i-1).getNodeNumber()+ " to node "+ route.get(i).getNodeNumber()+" :"+ route.get(i-1).getSailingTime((route.get(i))));
			cost += route.get(i-1).getSailingTime((route.get(i)))*load*loadCoef;
			System.out.println("Cost when in node "+route.get(i).getNodeNumber()+ ": "+ cost);
			
			if(route.get(i) instanceof DeliveryNode){
				DeliveryNode d = (DeliveryNode) route.get(i);
				load = load - d.getDemand();
			}
			if(route.get(i) instanceof PickupNode){
				PickupNode p = (PickupNode) route.get(i);
				load = load + p.getBackload();
			}
			
		}
		System.out.println("End cost: "+ cost);
		return cost;
	}
	
}

